package org.gecko.logging.log4j;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;

import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.core.LogEvent;
import org.apache.logging.log4j.core.LoggerContext;
import org.apache.logging.log4j.core.impl.Log4jLogEvent;
import org.apache.logging.log4j.core.impl.Log4jLogEvent.Builder;
import org.apache.logging.log4j.core.impl.ThrowableProxy;
import org.apache.logging.log4j.message.Message;
import org.osgi.framework.Bundle;
import org.osgi.service.log.LogEntry;
import org.osgi.service.log.LogLevel;
import org.osgi.service.log.LogListener;
import org.osgi.service.log.admin.LoggerAdmin;

public class Log4JLogListener implements LogListener {//, LoggerContextListener {

    private static final String EVENTS_BUNDLE = "Events.Bundle";
    private static final String EVENTS_FRAMEWORK = "Events.Framework";
    private static final String EVENTS_SERVICE = "Events.Service";
    private static final String LOG_SERVICE = "LogService";

    volatile LoggerContext loggerContext;
    volatile Logger rootLogger;
    final Map<String, LogLevel> initialLogLevels;
    final org.osgi.service.log.admin.LoggerContext osgiLoggerContext;
	private LoggerContext context;

    public Log4JLogListener(LoggerAdmin loggerAdmin) {
        osgiLoggerContext = loggerAdmin.getLoggerContext(null);
        initialLogLevels = osgiLoggerContext.getLogLevels();

        context = (LoggerContext) LogManager.getContext(false);
        onStart(context);
    }

    @Override
    public void logged(final LogEntry entry) {
        String loggerName = entry.getLoggerName();
        String message = entry.getMessage();
        Object[] arguments = null;
        Level level = from(entry.getLogLevel());
        final AtomicBoolean avoidCallerData = new AtomicBoolean();

        if (EVENTS_BUNDLE.equals(loggerName) ||
            EVENTS_FRAMEWORK.equals(loggerName) ||
            LOG_SERVICE.equals(loggerName)) {

            loggerName = formatBundle(entry.getBundle(), loggerName);
            avoidCallerData.set(true);
        }
        else if (EVENTS_SERVICE.equals(loggerName)) {
            loggerName = formatBundle(entry.getBundle(), loggerName);
            message = message + " {}";
            arguments = new Object[] {entry.getServiceReference()};
            avoidCallerData.set(true);
        }

        Logger logger = loggerContext.getLogger(loggerName);

        // Check to see if there's a logger defined in our configuration and
        // if there is, then make sure it's handled as an override for the
        // effective level.
        if (!logger.equals(rootLogger) && !logger.isEnabled(level)) {
            return;
        }
        
        Message msg = rootLogger.getMessageFactory().newMessage(message, arguments);
        
        Builder eventBuilder = Log4jLogEvent.newBuilder();
        LogEvent event = eventBuilder.setLevel(level)
        	.setLoggerName(loggerName)
        	.setThreadName(entry.getThreadInfo())
        	.setThrownProxy(getThrowableProxy(entry.getException()))
        	.setSource(entry.getLocation())
        	.setTimeMillis(entry.getTime())
        	.setMessage(msg).build();
        loggerContext.getConfiguration().getRootLogger().log(event);
    }

    public void onStart(LoggerContext context) {
        loggerContext = context;
        loggerContext.getRootLogger();
        rootLogger = loggerContext.getRootLogger();

        Map<String, LogLevel> updatedLevels = updateLevels(loggerContext, initialLogLevels);
        osgiLoggerContext.setLogLevels(updatedLevels);
    }

    String formatBundle(Bundle bundle, String loggerName) {
        return new StringBuilder().append(
            loggerName
        ).append(
            "."
        ).append(
            bundle.getSymbolicName()
        ).toString();
    }

    LogLevel from(Level level) {
        if (Level.ALL.equals(level)) {
            return LogLevel.TRACE;
        }
        else if (Level.DEBUG.equals(level)) {
            return LogLevel.DEBUG;
        }
        else if (Level.ERROR.equals(level)) {
            return LogLevel.ERROR;
        }
        else if (Level.INFO.equals(level)) {
            return LogLevel.INFO;
        }
        else if (Level.TRACE.equals(level)) {
            return LogLevel.TRACE;
        }
        else if (Level.WARN.equals(level)) {
            return LogLevel.WARN;
        }

        return LogLevel.WARN;
    }

    Level from(LogLevel logLevel) {
        switch (logLevel) {
            case AUDIT:
                return Level.TRACE;
            case DEBUG:
                return Level.DEBUG;
            case ERROR:
                return Level.ERROR;
            case INFO:
                return Level.INFO;
            case TRACE:
                return Level.TRACE;
            case WARN:
            default:
                return Level.WARN;
        }
    }

    ThrowableProxy getThrowableProxy(Throwable t) {
        if (t == null)
            return null;

        ThrowableProxy throwableProxy = new ThrowableProxy(t);
        return throwableProxy;
    }

    Map<String, LogLevel> updateLevels(LoggerContext loggerContext, Map<String, LogLevel> levels) {
        Map<String, LogLevel> copy = new HashMap<String, LogLevel>(levels);

        Logger root = loggerContext.getRootLogger();
        LogLevel rootLevel = from(root.getLevel());
        copy.put(org.osgi.service.log.Logger.ROOT_LOGGER_NAME, rootLevel);
        copy.put(EVENTS_BUNDLE, LogLevel.TRACE);
        copy.put(EVENTS_FRAMEWORK, LogLevel.TRACE);
        copy.put(EVENTS_SERVICE, LogLevel.TRACE);
        copy.put(LOG_SERVICE, LogLevel.TRACE);

        for (Logger logger : loggerContext.getLoggers()) {
            String name = logger.getName();
            Level level = logger.getLevel();

            if (level != null) {
                copy.remove(name);

                if (level != Level.OFF) {
                    copy.put(name, from(level));
                }
            }
        }

        return copy;
    }

}
